// Filename :   FFT.H             
// Author   :   Jonathan Valvano           
// Date     :   Feb 12, 2008               
// derived from Numerical Recipes in C, Cambridge University Press

//#include <stdlib.h>
//#include <math.h>
#include <cstdlib>
#include <cmath>



#define SWAP(a,b) tempr = (a); (a) = (b); (b) = tempr

// Input: nn is the number of points in the data and in the FFT, 
//           nn must be a power of 2
// Input: data is sampled voltage v(0),0,v(1),0,v(2),...v(nn-1),0 versus time
// Output: data is complex FFT Re[V(0)],Im[V(0)], Re[V(1)],Im[V(1)],...
// data is an array of 2*nn elements
void fft(long double data[], unsigned long nn);

//-----------------------------------------------------------
// Calculates the FFT magnitude at a given frequency index. 
// Input: data is complex FFT Re[V(0)],Im[V(0)], Re[V(1)],Im[V(1)],...
// Input: nn is the number of points in the data and in the FFT, 
//           nn must be a power of 2
// Input: k is frequency index 0 to nn/2-1
//        E.g., if nn=16384, then k can be 0 to 8191
// Output: Magnitude in volts at this frequency (volts)
// data is an array of 2*nn elements
// returns 0 if k >= nn/2
long double fftMagnitude(long double data[], unsigned long nn, unsigned long k);

//-----------------------------------------------------------
// Calculates the FFT magnitude in db full scale at a given frequency index. 
// Input: data is complex FFT Re[V(0)],Im[V(0)], Re[V(1)],Im[V(1)],...
// Input: nn is the number of points in the data and in the FFT, 
//           nn must be a power of 2
// Input: k is frequency index 0 to nn/2-1
//        E.g., if nn=16384, then k can be 0 to 8191
// Input: fullScale is the largest possible component in volts
// Output: Magnitude in db full scale at this frequency
// data is an array of 2*nn elements
// returns -200 if k >= nn/2
long double fftMagdB(long double data[], unsigned long nn, unsigned long k, double fullScale);

//-----------------------------------------------------------
// Calculates the FFT phase at a given frequency index. 
// Input: data is complex FFT Re[V(0)],Im[V(0)], Re[V(1)],Im[V(1)],...
// Input: nn is the number of points in the data and in the FFT, 
//           nn must be a power of 2
// Input: k is frequency index 0 to nn/2-1
//        E.g., if nn=16384, then k can be 0 to 8191
// Output: Phase at this frequency
// data is an array of 2*nn elements
// returns 0 if k >= nn/2
long double fftPhase(long double data[], unsigned long nn, unsigned long k);

//-----------------------------------------------------------
// Calculates equivalent frequency in Hz at a given frequency index. 
// Input: fs is sampling rate in Hz
// Input: nn is the number of points in the data and in the FFT, 
//           nn must be a power of 2
// Input: k is frequency index 0 to nn-1
//        E.g., if nn=16384, then k can be 0 to 16383
// Output: Equivalent frequency in Hz
// returns 0 if k >= nn
long double fftFrequency (unsigned long nn, unsigned long k, long double fs);
